/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2012, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthm@freeswitch.org>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Luis Azedo <luis@2600hz.com>
 *
 * mod_presence.c -- presence custom handler
 *
 */
#include "mod_presence.h"
#include "hashes.h"

typedef int(*allowed_char_fun)(char);

static int is_allowed_user_char(char ch_uri_part)
{
	return ch_uri_part == '.'
		|| ch_uri_part == '*'
		|| ch_uri_part == '#'
		|| ch_uri_part == '_'
		|| ch_uri_part == '-'
		|| ch_uri_part == '+'
		|| isalnum(ch_uri_part);
}

static int is_allowed_domain_char(char ch_uri_part)
{
	return ch_uri_part == '.'
		|| ch_uri_part == '-'
		|| isalnum(ch_uri_part);
}

static int is_valid_uri_part(char* str_uri_part, allowed_char_fun fun)
{
	while(*str_uri_part) {
		if(!fun(*str_uri_part)) {
			return 0;
		}
		str_uri_part++;
	}
	return 1;
}

static int is_uri_user_valid(char* str_uri_part)
{
	return is_valid_uri_part(str_uri_part, is_allowed_user_char);
}

static int is_uri_domain_valid(char* str_uri_part)
{
	return is_valid_uri_part(str_uri_part, is_allowed_domain_char);
}

static int is_ip(char* str_ip)
{
	struct addrinfo hint, *res = NULL;
	int ret = 0;

	memset(&hint, 0, sizeof(hint));

	hint.ai_family = PF_UNSPEC;
	hint.ai_flags = AI_NUMERICHOST;

	ret = getaddrinfo(str_ip, NULL, &hint, &res);
	if (ret) {
		return 0;
	}
	ret = 0;

	if(res->ai_family == AF_INET) {
		ret = 1;
	} else if (res->ai_family == AF_INET6) {
		ret = 1;
	}
	freeaddrinfo(res);
	return ret;
}

static const char *realm_list[] = {
	"variable_ecallmgr_Account-Realm",
	"variable_ecallmgr_Realm",
	"variable_sip_to_host",
	"variable_sip_invite_domain",
	"variable_domain_name",
	NULL
};

static char* pres_event_get_first_of(switch_event_t *event, const char *list[])
{
	switch_event_header_t *header = NULL;
	int i = 0;
	while(list[i] != NULL) {
		if((header = switch_event_get_header_ptr(event, list[i])) != NULL)
			break;
		i++;
	}
	if(header != NULL) {
		return header->value;
	} else {
		return "nodomain";
	}
}

static char* kz_presence_direction_in(switch_event_t *event)
{
	char *direction;
	if ((direction = switch_event_get_header(event, "Presence-Call-Direction")) == NULL)
		direction = switch_event_get_header_nil(event, "presence-call-direction");
	return direction;
}

static char* kz_presence_to_initiator(switch_event_t *event)
{
	char *to;
	if ((to = switch_event_get_header(event, "Caller-Destination-Number")) == NULL)
		to = switch_event_get_header_nil(event, "variable_sip_to_user");
	return to;
}

static char* kz_presence_to_recipient(switch_event_t *event)
{
	char *to;
	if ((to = switch_event_get_header(event, "Caller-Caller-ID-Number")) == NULL)
		to = switch_event_get_header_nil(event, "variable_sip_from_user");
	return to;
}

static char* kz_presence_to(char *direction, switch_event_t *event)
{
	if (!strcasecmp(direction, "initiator"))
		return kz_presence_to_initiator(event);

	return kz_presence_to_recipient(event);

}

static char* kz_presence_direction(char* direction)
{
	if (direction && !strcasecmp(direction, "inbound"))
		return "initiator";

	return "recipient";

}

static char* kz_presence_expires(char* state)
{
	if (!strcasecmp(state, "terminated"))
		return "20";

	return "0";
}

static char* kz_presence_state(char* direction, char* status, char* answer_state)
{
	if (!strcasecmp(answer_state, "answered"))
		return "confirmed";

	if (!strcasecmp(answer_state, "hangup"))
		return "terminated";

	if (!strcasecmp(status, "hangup"))
		return "terminated";

	if (!strcasecmp(status, "cs_hangup"))
		return "terminated";

	if (!strcasecmp(status, "cs_routing") && !strcasecmp(direction, "initiator"))
		return "confirmed";

	if (!strcasecmp(direction, "initiator") && !strcasecmp(answer_state, "ringing"))
		return "confirmed";

	if (!strcasecmp(direction, "recipient") && !strcasecmp(answer_state, "ringing"))
		return "early";

	return answer_state;

}

static void presence_fire_event(switch_core_session_t *session, switch_event_t *event, char* presence_state)
{
	const char *uuid = switch_event_get_header_nil(event, "Unique-ID");
	char *presence_expires = kz_presence_expires(presence_state);
	switch_event_t *pres_event;
	switch_channel_t *channel = switch_core_session_get_channel(session);

	char *presence_user = (char*) switch_channel_get_private(channel, "__kazoo_presence_user__");
	char *presence_realm = (char*) switch_channel_get_private(channel, "__kazoo_presence_realm__");
	char *presence_direction = (char*) switch_channel_get_private(channel, "__kazoo_presence_direction__");
	char *to = (char*) switch_channel_get_private(channel, "__kazoo_presence_to__");

	if (switch_event_create_subclass(&pres_event, SWITCH_EVENT_CUSTOM, MY_EVENT_PRESENCE_DIALOG) == SWITCH_STATUS_SUCCESS) {
		switch_event_merge(pres_event, event);
		pres_event->flags |= EF_UNIQ_HEADERS;
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Direction", presence_direction);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-State", presence_state);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Expires", presence_expires);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Realm", presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Call-ID-Hash", "%u", core_hash(uuid, strlen(uuid), 0));
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Realm-Hash", "%u", core_hash(presence_realm, strlen(presence_realm), 0));
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Presence-ID", "%s@%s", presence_user, presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Presence-URI", "sip:%s@%s", presence_user, presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-From", "sip:%s@%s", presence_user, presence_realm);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-From-User", presence_user);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-To", "sip:%s@%s", to, presence_realm);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-To-User", to);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Dialog-Event-Package", "dialog");

		// compatibility
		if(!strcmp(presence_direction, "initiator")) {
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Caller-Destination-Number", to);
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "variable_sip_to_user", to);
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Call-Direction", "inbound");
		} else {
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Caller-Caller-ID-Number", to);
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "variable_sip_from_user", to);
			switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "Presence-Call-Direction", "outbound");
		}
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "variable_ecallmgr_Realm", presence_realm);
		switch_event_add_header_string(pres_event, SWITCH_STACK_BOTTOM, "variable_sip_invite_domain", presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "from", "%s@%s", presence_user, presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "variable_presence_id", "%s@%s", presence_user, presence_realm);
		switch_event_add_header(pres_event, SWITCH_STACK_BOTTOM, "Channel-Presence-ID", "%s@%s", presence_user, presence_realm);

		switch_event_fire(&pres_event);
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "fired new presence state %s for %s@%s\n", presence_state, presence_user, presence_realm);
	} else {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_ERROR, "Couldn't create event to fire new state %s for %s@%s\n", presence_state, presence_user, presence_realm);
	}
}

static switch_status_t kz_presence_on_hangup(switch_core_session_t *session)
{
	switch_channel_t *channel = switch_core_session_get_channel(session);
	switch_event_t *event;
	char* uuid = switch_core_session_get_uuid(session);
	char *presence_user = (char*) switch_channel_get_private(channel, "__kazoo_presence_user__");
	char *presence_realm = (char*) switch_channel_get_private(channel, "__kazoo_presence_realm__");

	switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "ON HANGUP for %s@%s\n", presence_user, presence_realm);

	if(presence_user && presence_realm) {
		if (switch_event_create(&event, SWITCH_EVENT_PRESENCE_IN) == SWITCH_STATUS_SUCCESS) {
			switch_channel_event_set_data(channel, event);
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "proto", "any");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "rpid", "unknown");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "status", "CS_HANGUP");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "event_type", "presence");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "alt_event_type", "dialog");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "presence-call-info-state", "idle");
			switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "presence-call-direction",
										   switch_channel_direction(channel) == SWITCH_CALL_DIRECTION_OUTBOUND ? "outbound" : "inbound");
			if (switch_true(switch_channel_get_variable(channel, "presence_privacy"))) {
				switch_event_add_header_string(event, SWITCH_STACK_BOTTOM, "Presence-Privacy", "true");
			}
			presence_fire_event(session, event, "terminated");
			switch_event_destroy(&event);
		} else {
			switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_ERROR, "Couldn't create event to fire new state terminated for %s@%s\n", presence_user, presence_realm);
		}
	}

	return SWITCH_STATUS_SUCCESS;
}

static int presence_update_state(switch_core_session_t *session, switch_event_t *event, char *presence_user, char *presence_realm, char* presence_direction, char* presence_state)
{
	char *to = kz_presence_to(presence_direction, event);
	int result = 0;
	const char *channel_call_info_state = NULL;
	switch_channel_t *channel = switch_core_session_get_channel(session);
	channel_call_info_state = (char*) switch_channel_get_private(channel, "__kazoo_presence_state__");
	if (channel_call_info_state == NULL) {
		switch_channel_set_private(channel, "__kazoo_presence_state__", switch_core_session_strdup(session, presence_state));
		switch_channel_set_private(channel, "__kazoo_presence_user__", switch_core_session_strdup(session, presence_user));
		switch_channel_set_private(channel, "__kazoo_presence_realm__", switch_core_session_strdup(session, presence_realm));
		switch_channel_set_private(channel, "__kazoo_presence_direction__", switch_core_session_strdup(session, presence_direction));
		switch_channel_set_private(channel, "__kazoo_presence_to__", switch_core_session_strdup(session, to));
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(session), SWITCH_LOG_DEBUG1, "created presence channel state %s for %s@%s\n", presence_state, presence_user, presence_realm);
		result = 1;
	} else if(!strcasecmp(presence_state, channel_call_info_state)) {
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(session), SWITCH_LOG_DEBUG1, "SKIPPING channel state %s : %s\n", channel_call_info_state, switch_channel_get_uuid(channel));
	} else {
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(session), SWITCH_LOG_DEBUG1, "updated presence channel state %s => %s for %s@%s\n", presence_state, channel_call_info_state, presence_user, presence_realm);
		switch_channel_set_private(channel, "__kazoo_presence_state__", switch_core_session_strdup(session, presence_state));
		result = 1;
	}
	return result;
}

static char* extract_presence_id(switch_event_t *event, char** presence_user, char** presence_realm)
{
	const char *uuid = switch_event_get_header_nil(event, "Unique-ID");
	char *presence_id = switch_event_get_header(event, "variable_presence_id");
	char *presence_user_tmp = NULL, *presence_realm_tmp = NULL;
	char *presence_tmp = NULL;

	presence_tmp = switch_safe_strdup(presence_id);
	switch_split_user_domain(presence_tmp, &presence_user_tmp, &presence_realm_tmp);
	if (zstr(presence_realm_tmp) || is_ip(presence_realm_tmp)) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "REALM for %s is NULL or IP Address, getting realm from default variables\n", presence_id);
		presence_realm_tmp = pres_event_get_first_of(event, realm_list);
	}
	if(zstr(presence_realm_tmp)) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "SKIPPING NULL REALM for %s\n", presence_id);
		switch_safe_free(presence_tmp);
		return NULL;
	}
	if(is_ip(presence_realm_tmp)) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "SKIPPING IP REALM %s for %s\n", presence_realm_tmp, presence_id);
		switch_safe_free(presence_tmp);
		return NULL;
	}
	if(!is_uri_user_valid(presence_user_tmp) || !is_uri_domain_valid(presence_realm_tmp)) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "SKIPPING INVALID PRESENTITY [sip:%s@%s]\n", presence_user_tmp, presence_realm_tmp);
		switch_safe_free(presence_tmp);
		return NULL;
	}

	*presence_user = presence_user_tmp;
	*presence_realm = presence_realm_tmp;
	// this must be freed
	return presence_tmp;

}

static char* get_presence_id(switch_core_session_t *session, switch_event_t *event, char** presence_user, char** presence_realm)
{
	char *presence_id = NULL;
	const char *uuid = switch_event_get_header_nil(event, "Unique-ID");
	switch_channel_t *channel = switch_core_session_get_channel(session);

	*presence_user = NULL;
	*presence_realm = NULL;

	if(switch_channel_get_private(channel, "__kazoo_no_presence__") != NULL) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "__kazoo_no_presence__ IS SET\n");
		return NULL;
	}

	*presence_user = (char*) switch_channel_get_private(channel, "__kazoo_presence_user__");
	*presence_realm = (char*) switch_channel_get_private(channel, "__kazoo_presence_realm__");

	if(*presence_user && *presence_realm) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "presence_user && presence_realm NOT NULL\n");
		return NULL;
	}

	presence_id = extract_presence_id(event, presence_user, presence_realm);
	if(presence_id == NULL) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "SETTING __kazoo_no_presence__");
		switch_channel_set_private(channel, "__kazoo_no_presence__", switch_core_session_strdup(session, "true"));
	}

	return presence_id;
}

static void presence_event_handler(switch_event_t *event)
{
	switch_core_session_t *session = NULL;
	const char *uuid = switch_event_get_header_nil(event, "Unique-ID");
	char *proto = switch_event_get_header_nil(event, "proto");
	char *presence_id = NULL, *presence_user = NULL, *presence_realm = NULL;
	char *status = switch_event_get_header_nil(event, "status");
	char *answer_state = switch_event_get_header_nil(event, "Answer-State");
	char *presence_direction = kz_presence_direction(kz_presence_direction_in(event));
	char *presence_state = kz_presence_state(presence_direction, status, answer_state);

	if (!strcmp(presence_state, "terminated")) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "SKIPPING terminated\n");
		// processed in kz_presence_on_hangup
		return;
	}

	if (zstr(proto) || strcasecmp(proto, "any")) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "SKIPPING proto %s for %s@%s\n", proto, presence_user, presence_realm);
		return;
	}

	switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG, "presence ( %s , %s , %s ) => %s\n", presence_direction, status, answer_state, presence_state);

	if ((session = switch_core_session_locate(uuid))) {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "GOT SESSION\n");
		presence_id = get_presence_id(session, event, &presence_user, &presence_realm);
		if(presence_user && presence_realm) {
			switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "updating presence %s@%s => %s\n", presence_user, presence_realm, presence_state);
			if(presence_update_state(session, event, presence_user, presence_realm, presence_direction, presence_state) == 1) {
				switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "firing event presence %s@%s => %s\n", presence_user, presence_realm, presence_state);
				presence_fire_event(session, event, presence_state);
			}
		} else {
			switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_DEBUG1, "NO presence_user && presence_realm\n");
		}

		switch_safe_free(presence_id)
		switch_core_session_rwunlock(session);
	} else {
		switch_log_printf(SWITCH_CHANNEL_UUID_LOG(uuid), SWITCH_LOG_ERROR, "Couldn't find session to process presence state\n");
	}
}


static switch_state_handler_table_t kz_presence_state_handlers = {
	/*.on_init */ NULL,
	/*.on_routing */ NULL,
	/*.on_execute */ NULL,
	/*.on_hangup */ kz_presence_on_hangup,
	/*.on_exchange_media */ NULL,
	/*.on_soft_execute */ NULL,
	/*.on_consume_media */ NULL,
	/*.on_hibernate */ NULL,
	/*.on_reset */ NULL,
	/*.on_park */ NULL,
	/*.on_reporting */ NULL,
	/*on.destroy */ NULL,
	SSH_FLAG_PRE_EXEC
};

static void kz_presence_register_state_handlers()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "registering state handler\n");
	switch_core_add_state_handler(&kz_presence_state_handlers);
}

static void kz_presence_unregister_state_handlers()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "unregistering state handler\n");
	switch_core_remove_state_handler(&kz_presence_state_handlers);
}

static void kz_presence_bind_events()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "binding for events\n");
	if (switch_event_bind("kz_presence", SWITCH_EVENT_PRESENCE_IN, SWITCH_EVENT_SUBCLASS_ANY, presence_event_handler, NULL) != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Couldn't bind to presence_in event!\n");
	}
}

static void kz_presence_unbind_events()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "unbinding events\n");
	switch_event_unbind_callback(presence_event_handler);
}

static void kz_presence_register_events()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "registering custom event\n");
	if (switch_event_reserve_subclass(MY_EVENT_PRESENCE_DIALOG) != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Couldn't register subclass %s!\n", MY_EVENT_PRESENCE_DIALOG);
	}
}

static void kz_presence_unregister_events()
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "unregistering custom event\n");
	switch_event_free_subclass(MY_EVENT_PRESENCE_DIALOG);
}

void kz_presence_start()
{
	kz_presence_register_events();
	kz_presence_bind_events();
	kz_presence_register_state_handlers();
}

void kz_presence_stop()
{
	kz_presence_unregister_state_handlers();
	kz_presence_unbind_events();
	kz_presence_unregister_events();
}

SWITCH_MODULE_LOAD_FUNCTION(mod_presence_load);
SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_presence_shutdown);
SWITCH_MODULE_DEFINITION(mod_presence,mod_presence_load, mod_presence_shutdown, NULL);

SWITCH_MODULE_LOAD_FUNCTION(mod_presence_load) {
	*module_interface = switch_loadable_module_create_module_interface(pool, modname);
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "loading mod_presence\n");
	kz_presence_start();

	/* indicate that the module should continue to be loaded */
	return SWITCH_STATUS_SUCCESS;
}

SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_presence_shutdown) {

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "unloading mod_presence\n");
	kz_presence_stop();

    return SWITCH_STATUS_SUCCESS;
}

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4:
 */

